// -*- coding: iso-8859-1-unix -*-

/*
** ComponentContext.cpp
*/

#include <scs/ComponentContext.hpp>
#include <scs/IComponent.hpp>
#include <scs/IReceptacles.hpp>
#include <scs/IMetaInterface.hpp>
#include <scs/SCSException.hpp>

#include <CORBA.h>
#include <memory>

namespace scs {
  namespace core {

    class component_context_access
    {
    public:
      static void CreateBasicFacets(ComponentContext& self)
      {
        std::auto_ptr<IComponentImpl> component(new IComponentImpl(self));
        std::auto_ptr<IReceptaclesImpl> receptacles(new IReceptaclesImpl(self));
        std::auto_ptr<IMetaInterfaceImpl> meta_interface(new IMetaInterfaceImpl(self));

        {
          PortableServer::ServantBase* servant = component.get();
          self.component_object_id = self.poa->activate_object(component.get());
          component.release()->_remove_ref(); // We remove the implicit reference we're holding
          CORBA::Object_var obj = self.poa->id_to_reference(self.component_object_id);
          assert(!CORBA::is_nil(obj)); // it was already activated, must not be nil
          self.component_reference = IComponent::_narrow(obj);
          assert(!CORBA::is_nil(self.component_reference)); // We know it is a IComponent, must not be nil

          self.facets_by_interface.insert(std::pair<std::string, CORBA::Object_var>
                                          ("IDL:scs/core/IComponent:1.0", obj));
          Facet facet = {"IComponent", "IDL:scs/core/IComponent:1.0", obj, servant};
          self.facets_by_name.insert(std::pair<std::string, Facet>("IComponent", facet));
        }

        {
          PortableServer::ServantBase* servant = receptacles.get();
          self.receptacles_object_id = self.poa->activate_object(receptacles.get());
          receptacles.release()->_remove_ref(); // We remove the implicit reference we're holding
          CORBA::Object_var obj = self.poa->id_to_reference(self.receptacles_object_id);
          assert(!CORBA::is_nil(obj)); // it was already activated, must not be nil
          self.receptacles_reference = IReceptacles::_narrow(obj);
          assert(!CORBA::is_nil(self.receptacles_reference)); // We know it is a IComponent, must not be nil
          
          self.facets_by_interface.insert(std::pair<std::string, CORBA::Object_var>
                                          ("IDL:scs/core/IReceptacles:1.0", obj));
          Facet facet = {"IReceptacles", "IDL:scs/core/IReceptacles:1.0", obj, servant};
          self.facets_by_name.insert(std::pair<std::string, Facet>("IReceptacles", facet));
        }

        {
          PortableServer::ServantBase* servant = meta_interface.get();
          self.meta_interface_object_id = self.poa->activate_object(meta_interface.get());
          meta_interface.release()->_remove_ref(); // We remove the implicit reference we're holding
          CORBA::Object_var obj = self.poa->id_to_reference(self.meta_interface_object_id);
          assert(!CORBA::is_nil(obj)); // it was already activated, must not be nil
          self.meta_interface_reference = IMetaInterface::_narrow(obj);
          assert(!CORBA::is_nil(self.meta_interface_reference)); // We know it is a IComponent, must not be nil
          
          self.facets_by_interface.insert(std::pair<std::string, CORBA::Object_var>
                                          ("IDL:scs/core/IMetaInterface:1.0", obj));
          Facet facet = {"IMetaInterface", "IDL:scs/core/IMetaInterface:1.0", obj, servant};
          self.facets_by_name.insert(std::pair<std::string, Facet>("IMetaInterface", facet));
        }
      }
    };

    ComponentContext::ComponentContext(CORBA::ORB_var orb, ComponentId const& id)
      : component_id(id)
    {
      CORBA::Object_var poa_obj = orb->resolve_initial_references("RootPOA");
      poa = PortableServer::POA::_narrow(poa_obj);
      
      scs::core::component_context_access::CreateBasicFacets(*this);
    }

    ComponentContext::ComponentContext(CORBA::ORB_var orb, PortableServer::POA_var poa, ComponentId const& id)
      : component_id(id), poa(poa)
    {
      scs::core::component_context_access::CreateBasicFacets(*this);
    }

    ComponentContext::~ComponentContext()
    {
      try
      {
        for(std::map<std::string, Facet>::iterator
              first = facets_by_name.begin()
              , last = facets_by_name.end()
              ;first != last; ++first)
        {
          poa->deactivate_object(*poa->reference_to_id(first->second.reference));
        }
      }
      catch(...) {}
    }

    ComponentId ComponentContext::getComponentId() 
    {
      return component_id;
    }

    scs::core::IComponent_var ComponentContext::getIComponent() 
    {
      return component_reference;
    }

    void ComponentContext::addFacet(const char* name, const char* interface_name, PortableServer::ServantBase* s)
    {
      std::auto_ptr<PortableServer::ServantBase> p(s);
      addFacet(name, interface_name, p);
    }

    void ComponentContext::updateFacet(const char* name, PortableServer::ServantBase* s)
    {
      std::auto_ptr<PortableServer::ServantBase> p(s);
      updateFacet(name, p);
    }

    struct deactivate_if_not_commit
    {
      deactivate_if_not_commit(PortableServer::POA_var& poa
                               , PortableServer::ObjectId_var& id)
        : commit(false), poa(poa), id(id) {}
      ~deactivate_if_not_commit()
      {
        if(!commit)
          poa->deactivate_object(*id);
      }
      bool commit;
      PortableServer::POA_var& poa;
      PortableServer::ObjectId_var& id;
    };

    void ComponentContext::addFacet(const char* name, const char* interface_name
                                    , std::auto_ptr<PortableServer::ServantBase>& s)
    {
#ifdef SCS_THREADING_ENABLED
      facet_lock l(*this);
#endif
      addFacet_internal(name, interface_name, s);
    }

    void ComponentContext::addFacet_internal(const char* name, const char* interface_name
                                    , std::auto_ptr<PortableServer::ServantBase>& s)
    {
      PortableServer::ObjectId_var id = poa->activate_object(s.get());

      deactivate_if_not_commit commit_(poa, id);

      CORBA::Object_var obj = poa->id_to_reference(id);      
      assert(!CORBA::is_nil(obj)); // it was already activated, must not be nil
      if(facets_by_name.find(name) == facets_by_name.end())
      {
        facets_by_interface.insert(std::pair<std::string, CORBA::Object_var>
                                   (interface_name, obj));
        Facet facet = {name, interface_name, obj, s.get()};
        facets_by_name.insert(std::pair<std::string, Facet>(name, facet));
        commit_.commit = true;
        s.release();
      }
      else
        throw SCSException("Facet already exists");
    }

    void ComponentContext::updateFacet(const char* name, std::auto_ptr<PortableServer::ServantBase>& s)
    {
#ifdef SCS_THREADING_ENABLED
      facet_lock l(*this);
#endif
      std::string interfaceName;

      std::map<std::string, Facet>::iterator
        iterator = facets_by_name.find(name);
      assert(iterator != facets_by_name.end());
      if(iterator != facets_by_name.end())
      {
        interfaceName = iterator->second.interfaceName;
      }
      else
        throw SCSException("Facet doesn't exist");
      
      removeFacet_internal(name);
      addFacet_internal(name, interfaceName.c_str(), s);
    }

    void ComponentContext::addReceptacle(const char* name, const char* interface_name, bool multiplexed)
    {
#ifdef SCS_THREADING_ENABLED
      receptacles_lock l(*this);
#endif
      receptacles.insert(std::pair<std::string, Receptacle>(name, Receptacle(name, interface_name, multiplexed)));
    }

    void ComponentContext::deactivateFacet_internal(const char* name)
    {
      std::map<std::string, Facet>::iterator
        iterator = facets_by_name.find(name);
      assert(iterator != facets_by_name.end());
      if(iterator != facets_by_name.end())
      {
        poa->deactivate_object(*poa->reference_to_id(iterator->second.reference));
      }
      else
        throw SCSException("Facet doesn't exist");
    }

    void ComponentContext::deactivateFacet(const char* name)
    {
#ifdef SCS_THREADING_ENABLED
      facet_lock l(*this);
#endif
      deactivateFacet_internal(name);
    }

    void ComponentContext::removeFacet_internal(const char* name)
    {
      deactivateFacet_internal(name);
      std::map<std::string, Facet>::iterator
        iterator = facets_by_name.find(name);
      assert(iterator != facets_by_name.end());
      if(iterator != facets_by_name.end())
      {
        facets_by_interface.erase(iterator->second.interfaceName);
        facets_by_name.erase(iterator);
      }
      else
        throw SCSException("Facet doesn't exist");
    }

    void ComponentContext::removeFacet(const char* name)
    {
#ifdef SCS_THREADING_ENABLED
      facet_lock l(*this);
#endif
      removeFacet_internal(name);
    }

    Facet ComponentContext::getFacetByName(const char* name)
    {
#ifdef SCS_THREADING_ENABLED
      facet_lock l(*this);
#endif
      std::map<std::string, Facet>::iterator iterator = facets_by_name.find(name);
      if(iterator != facets_by_name.end())
        return iterator->second;
      else
        throw SCSException("Facet doesn't exist");
    }
  }
}
