// -*- coding: iso-8859-1-unix -*-

/*
** IMetaInterface.cpp
*/

#include <scs/IMetaInterface.hpp>

#include <sstream>
#include <cstdlib>

namespace scs {
  namespace core {
    IMetaInterfaceImpl::IMetaInterfaceImpl(ComponentContext& context)
      : component(&context)
    {
    }

    IMetaInterfaceImpl::~IMetaInterfaceImpl() {
    }

    // O(n) Where n is the size of the facets map
    FacetDescriptions* IMetaInterfaceImpl::getFacets() 
    {
#ifdef SCS_THREADING_ENABLED
      ComponentContext::facet_lock lock(*component);
#endif
      std::map<std::string, Facet>const&
        map = component->getMapFacetsByName();
      FacetDescriptions_var descriptions(new FacetDescriptions(map.size()));
      descriptions->length(map.size());
      CORBA::ULong i = 0;
      for(std::map<std::string, Facet>
            ::const_iterator Iter = map.begin(), EndIter = map.end()
            ;Iter != EndIter;++Iter, ++i)
      {
        FacetDescription d;
        d.name = CORBA::string_dup(Iter->first.c_str());
        d.interface_name = CORBA::string_dup(Iter->second.interfaceName.c_str());
        d.facet_ref = Iter->second.reference;
        (*descriptions)[i] = d;
      }
      return descriptions._retn();
    }

    // O(log m*n) comparisons, where m is the size of the facets map
    // and n is the size of the NameList
    FacetDescriptions* IMetaInterfaceImpl::getFacetsByName(
      const NameList&  names)
    {
#ifdef SCS_THREADING_ENABLED
      ComponentContext::facet_lock lock(*component);
#endif
      std::map<std::string, Facet>const&
        map = component->getMapFacetsByName();
      FacetDescriptions_var descriptions(new FacetDescriptions);
      CORBA::ULong names_size = names.length();
      for(CORBA::ULong i = 0; i != names_size; ++i)
      {
        std::map<std::string, Facet>
          ::const_iterator Iter = map.find(names[i].in());
        if(Iter != map.end())
        {
          FacetDescription d;
          d.name = CORBA::string_dup(Iter->first.c_str());
          d.interface_name = CORBA::string_dup(Iter->second.interfaceName.c_str());
          d.facet_ref = Iter->second.reference;
          descriptions->length(i+1);
          (*descriptions)[i] = d;
        }
      }
      return descriptions._retn();
    }

    ReceptacleDescriptions* IMetaInterfaceImpl::getReceptacles()
    {
#ifdef SCS_THREADING_ENABLED
      ComponentContext::receptacles_lock lock(*component);
#endif
      std::map<std::string, Receptacle>const&
        map = component->getMapReceptacles();
      ReceptacleDescriptions_var descriptions(new ReceptacleDescriptions(map.size()));
      descriptions->length(map.size());
      CORBA::ULong i = 0;
      for(std::map<std::string, Receptacle>
            ::const_iterator Iter = map.begin(), EndIter = map.end()
            ;Iter != EndIter;++Iter, ++i)
      {
        ReceptacleDescription d;
        d.name = CORBA::string_dup(Iter->second.name.c_str());
        d.interface_name = CORBA::string_dup(Iter->second.interface_name.c_str());
        d.is_multiplex = Iter->second.multiplexed;
        (*descriptions)[i] = d;
        std::map<ConnectionId, CORBA::Object_var>const&
          connections = Iter->second.connections;
        (*descriptions)[i].connections.length(connections.size());
        std::size_t j = 0;
        for(std::map<ConnectionId, CORBA::Object_var>::const_iterator
              Iter = connections.begin()
              , EndIter = connections.end()
              ;Iter != EndIter
              ;++Iter, ++j)
        {
          ConnectionDescription cd;
          cd.id = Iter->first;
          cd.objref = Iter->second;
          (*descriptions)[i].connections[j] = cd;
        }
      }
      return descriptions._retn();
    }

    // O(log m*n) comparisons, where m is the size of the receptacles map
    // and n is the size of the NameList
    ReceptacleDescriptions* IMetaInterfaceImpl::getReceptaclesByName(
      const NameList &  names)
    {
#ifdef SCS_THREADING_ENABLED
      ComponentContext::receptacles_lock lock(*component);
#endif
      std::map<std::string, Receptacle>const&
        map = component->getMapReceptacles();
      ReceptacleDescriptions_var descriptions(new ReceptacleDescriptions(map.size()));
      CORBA::ULong i = 0;
      std::map<std::string, Receptacle>
        ::const_iterator Iter = map.begin(), EndIter = map.end();
      while(Iter != EndIter)
      {
        bool found = false;
        {
          for(int i = 0; i != names.length();++i)
          {
            CORBA::String_var string = names[i];
            if(Iter->second.name == string.in())
            {
              found = true;
              break;
            }
          }
        }
        if(found)
        {
          descriptions->length(map.size());
          ReceptacleDescription d;
          d.name = CORBA::string_dup(Iter->second.name.c_str());
          d.interface_name = CORBA::string_dup(Iter->second.interface_name.c_str());
          d.is_multiplex = Iter->second.multiplexed;
          (*descriptions)[i] = d;
          std::map<ConnectionId, CORBA::Object_var>const&
            connections = Iter->second.connections;
          (*descriptions)[i].connections.length(connections.size());
          std::size_t j = 0;
          for(std::map<ConnectionId, CORBA::Object_var>::const_iterator
                Iter = connections.begin()
                , EndIter = connections.end()
                ;Iter != EndIter
                ;++Iter, ++j)
          {
            ConnectionDescription cd;
            cd.id = Iter->first;
            cd.objref = Iter->second;
            (*descriptions)[i].connections[j] = cd;
          }
        }
        ++Iter; ++i;
      }
      return descriptions._retn();
    }
  }
}

